
from . import assert_processor_result
from pydoc_markdown.contrib.processors.pydocmd import PydocmdProcessor


def test_pydocmd_processor(processor=None):
  assert_processor_result(processor or PydocmdProcessor(),
  '''
  # Arguments
  s (str): A string.
  b (int): An int.
  ''',
  '''
  __Arguments__

  - __s__ (`str`): A string.
  - __b__ (`int`): An int.
  ''')

  # Intentionally not using the supplied processor as this is caught by
  # the Google processor.
  assert_processor_result(PydocmdProcessor(),
  '''
  Example:

  ```py
  scanner = ListScanner(lst)
  for value in scanner.safe_iter():
    if some_condition(value):
      value = scanner.advance()
  ```
  ''',
  '''
  Example:

  ```py
  scanner = ListScanner(lst)
  for value in scanner.safe_iter():
    if some_condition(value):
      value = scanner.advance()
  ```
  ''')



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Pydoc-markdown is an extensible framework for generating API documentation,
with a focus on Python source code and the Markdown output format.
"""


from nr.databind.core import Collect, Field, FieldName, ObjectMapper, Struct, UnionType
from nr.databind.json import JsonModule
from nr.stream import concat
from pydoc_markdown.interfaces import Context, Loader, Processor, Renderer, Resolver, Builder
from pydoc_markdown.contrib.loaders.python import PythonLoader
from pydoc_markdown.contrib.processors.filter import FilterProcessor
from pydoc_markdown.contrib.processors.crossref import CrossrefProcessor
from pydoc_markdown.contrib.processors.smart import SmartProcessor
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.util import ytemplate
from typing import List, Union
import docspec
import logging
import os
import subprocess
import yaml

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '3.3.1'

mapper = ObjectMapper(JsonModule())
logger = logging.getLogger(__name__)


class PydocMarkdown(Struct):
  """
  This object represents the main configuration for Pydoc-Markdown.
  """

  #: A list of loader implementations that load #docspec.Module#s.
  #: Defaults to #PythonLoader.
  loaders = Field([Loader], default=lambda: [PythonLoader()])

  #: A list of processor implementations that modify #docspec.Module#s. Defaults
  #: to #FilterProcessor, #SmartProcessor and #CrossrefProcessor.
  processors = Field([Processor], default=lambda: [
    FilterProcessor(), SmartProcessor(), CrossrefProcessor()])

  #: A renderer for #docspec.Module#s. Defaults to #MarkdownRenderer.
  renderer = Field(Renderer, default=MarkdownRenderer)

  #: Hooks that can be executed at certain points in the pipeline. The commands
  #: are executed with the current `SHELL`.
  hooks = Field({
    'pre_render': Field([str], FieldName('pre-render'), default=list),
    'post_render': Field([str], FieldName('post-render'), default=list),
  }, default=Field.DEFAULT_CONSTRUCT)

  # Hidden fields are filled at a later point in time and are not (de-) serialized.
  unknown_fields = Field([str], default=list, hidden=True)
  resolver = Field(Resolver, default=None, hidden=True)

  def __init__(self, *args, **kwargs) -> None:
    super(PydocMarkdown, self).__init__(*args, **kwargs)
    self.resolver = None
    self._context = None

  def load_config(self, data: Union[str, dict]) -> None:
    """
    Loads a YAML configuration from *data*.

    :param data: Nested structurre or the path to a YAML configuration file.
    """

    filename = None
    if isinstance(data, str):
      filename = data
      logger.info('Loading configuration file "%s".', filename)
      data = ytemplate.load(filename, {'env': ytemplate.Attributor(os.environ)})

    collector = Collect()
    result = mapper.deserialize(data, type(self), filename=filename, decorations=[collector])
    vars(self).update(vars(result))

    self.unknown_fields = list(concat((str(n.locator.append(u)) for u in n.unknowns)
      for n in collector.nodes))

  def init(self, context: Context) -> None:
    """
    Initialize all plugins with the specified *context*. Cannot be called multiple times.
    If omitted, the plugins will be initialized with a default context before the load,
    process or render phase.
    """

    if self._context:
      raise RuntimeError('already initialized')
    self._context = context
    logger.debug('Initializing plugins with context %r', context)
    for loader in self.loaders:
      loader.init(context)
    for processor in self.processors:
      processor.init(context)
    self.renderer.init(context)

  def ensure_initialized(self) -> None:
    if not self._context:
      self.init(Context(directory='.'))

  def load_modules(self) -> List[docspec.Module]:
    """
    Loads modules via the #loaders.
    """

    logger.info('Loading modules.')
    self.ensure_initialized()
    modules = []
    for loader in self.loaders:
      modules.extend(loader.load())
    return modules

  def process(self, modules: List[docspec.Module]) -> None:
    """
    Process modules via the #processors.
    """

    self.ensure_initialized()
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(modules)
    for processor in self.processors:
      processor.process(modules, self.resolver)

  def render(self, modules: List[docspec.Module], run_hooks: bool = True) -> None:
    """
    Render modules via the #renderer.
    """

    self.ensure_initialized()
    if run_hooks:
      self.run_hooks('pre-render')
    if self.resolver is None:
      self.resolver = self.renderer.get_resolver(modules)
    self.renderer.process(modules, self.resolver)
    self.renderer.render(modules)
    if run_hooks:
      self.run_hooks('post-render')

  def build(self, site_dir: str=None) -> None:
    if not Builder.provided_by(self.renderer):
      name = type(self.renderer).__name__
      raise NotImplementedError('Renderer "{}" does not support building'.format(name))
    self.ensure_initialized()
    self.renderer.build(site_dir)

  def run_hooks(self, hook_name: str) -> None:
    for command in getattr(self.hooks, hook_name.replace('-', '_')):
      subprocess.check_call(command, shell=True, cwd=self._context.directory)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor
import docspec


@implements(Processor)
class FilterProcessor(Struct):
  """
  The `filter` processor removes module and class members based on certain criteria.

  Example configuration:

  ```py
  - type: filter
    expression: not name.startswith('_') and default()
    documented_only: false
  ```

  ### Options
  """

  #: A Python expression that is evaluated given the variables `name`, `obj` and `default`
  #: and is expected to return a boolean to indicate whether the #docspec.ApiObject should
  #: be kept or removed. Default: `null`
  expression = Field(str, default=None)

  #: Keep only API objects that have docstrings. Default: `true`
  documented_only = Field(bool, default=True)

  #: Exclude API objects that appear to be private members (i.e. their name begins with
  #: and underscore but does not end with one). Default: `true`
  exclude_private = Field(bool, default=True)

  #: Exclude special members (e.g.` __path__`, `__annotations__`, `__name__` and `__all__`).
  #: Default: `true`
  exclude_special = Field(bool, default=True)

  #: Do not filter #docspec.Module objects. Default: `true`
  do_not_filter_modules = Field(bool, default=True)

  SPECIAL_MEMBERS = ('__path__', '__annotations__', '__name__', '__all__')

  @override
  def process(self, modules, _resolver):
    def m(obj):
      return self._match(obj)
    docspec.filter_visit(modules, m, order='post')

  def _match(self, obj: docspec.ApiObject) -> bool:
    if getattr(obj, 'members', []):
      return True
    if self.do_not_filter_modules and isinstance(obj, docspec.Module):
      return True
    if self.documented_only and not obj.docstring:
      return False
    if self.exclude_private and obj.name.startswith('_') and not obj.name.endswith('_'):
      return False
    if self.exclude_special and obj.name in self.SPECIAL_MEMBERS:
      return False
    if self.expression:
      scope = {'name': obj.name, 'obj': obj, 'default': _check}
      if not eval(self.expression, scope):  # pylint: disable=eval-used
        return False
    return True



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
This module defines the interfaces that can to be implemented for
Pydoc-Markdown to implement custom loaders for documentation data,
processors or renderers.
"""

from .util.databind import ChainTypeResolver
from nr.databind.core import  SerializeAs, UnionType
from nr.databind.core.union import EntrypointTypeResolver, ImportTypeResolver
from nr.interface import Interface, default
from typing import Iterable, List, Optional
import docspec
import subprocess


def _make_union_type(entrypoint_name: str) -> UnionType:
  return UnionType(ChainTypeResolver(
    EntrypointTypeResolver(entrypoint_name),
    ImportTypeResolver(),
  ))


class Context:
  """
  Context data that is passed to plugins when they are loaded.
  """

  def __init__(self, directory: str) -> None:
    self.directory = directory


class PluginBase(Interface):

  @default
  def init(self, context: Context) -> None:
    pass


@SerializeAs(_make_union_type('pydoc_markdown.interfaces.Loader'))
class Loader(PluginBase):
  """
  This interface describes an object that is capable of loading documentation
  data. The location from which the documentation is loaded must be defined
  with the configuration class.
  """

  def load(self) -> Iterable[docspec.Module]:
    """
    Load modules.
    """


class LoaderError(Exception):
  pass


class Resolver(Interface):
  """
  A resolver can be used by a #Processor to replace cross references with a hyperlink.
  """

  def resolve_ref(self, scope: docspec.ApiObject, ref: str) -> Optional[str]:
    pass


@SerializeAs(_make_union_type('pydoc_markdown.interfaces.Processor'))
class Processor(PluginBase):
  """
  A processor is an object that takes a list of #docspec.Module#s as an input and
  transforms it in an arbitrary way. This usually processes docstrings to convert from
  various documentation syntaxes to plain Markdown.
  """

  def process(self, modules: List[docspec.Module], resolver: Optional[Resolver]) -> None:
    pass


@SerializeAs(_make_union_type('pydoc_markdown.interfaces.Renderer'))
class Renderer(PluginBase):
  """
  A renderer is an object that takes a list of #docspec.Module#s as an input and produces
  output files or writes to stdout. It may also expose additional command-line arguments.
  There can only be one renderer at the end of the processor chain.

  Note that sometimes a renderer may need to perform some processing before the render step.
  To keep the possibility open that a renderer may implement generic processing that could
  used without the actual renderering functionality, #Renderer is a subclass of #Processor.
  """

  @default
  def process(self, modules: List[docspec.Module], resolver: Optional[Resolver]) -> None:
    pass

  @default
  def get_resolver(self, modules: List[docspec.Module]) -> Optional[Resolver]:
    return None

  def render(self, modules: List[docspec.Module]) -> None:
    pass


class Server(Interface):
  """
  This interface describes an object that can start a server process for
  live-viewing generated documentation in the browser. #Renderer
  implementations may additionally implement this interface to advocate their
  compatibility with the `--server` and `--open` options of the pydoc-markdown
  CLI.
  """

  def get_server_url(self) -> str:
    ...

  def start_server(self) -> subprocess.Popen:
    ...

  @default
  def reload_server(self, process: subprocess.Popen) -> subprocess.Popen:
    """
    Called when the files generated by pydoc-markdown have been updated.
    This gives the implementation a chance to reload the server process.
    The default implementation returns the *process* unchanged. Returning
    #None will automatically call #start_server() afterwards.
    """

    return process


class Builder(Interface):
  """
  This interface can be implemented additionally to the #Renderer interface to
  indicate that the renderer supports building another produce after the markdown
  files have been rendered.
  """

  def build(self, site_dir: str=None) -> None:
    """
    Invoke the build. If *site_dir* is specified, it is the directory in which the
    output files should be placed. Otherwise, the directory may be determined by
    the builder.
    """


@SerializeAs(_make_union_type('pydoc_markdown.interfaces.SourceLinker'))
class SourceLinker(PluginBase):
  """
  This interface is used to determine the URL to the source of an API object. Renderers
  can use it to place a link to the source in the generated documentation.
  """

  def get_source_url(self, obj: docspec.ApiObject) -> Optional[str]:
    pass




from . import assert_processor_result
from pydoc_markdown.contrib.processors.google import GoogleProcessor


def test_google_processor(processor=None):
  assert_processor_result(processor or GoogleProcessor(),
  '''
  Args:
    s (str): A string.
    b (int): An int.
  Returns:
    any: Something funny.
  ''',
  '''
  **Arguments**:

  - `s` _str_ - A string.
  - `b` _int_ - An int.

  **Returns**:

  - `any` - Something funny.
  ''')

  assert_processor_result(processor or GoogleProcessor(),
  '''
  Example:

  ```py
  scanner = ListScanner(lst)
  for value in scanner.safe_iter():
    if some_condition(value):
      value = scanner.advance()
  ```
  ''',
  '''
  **Example**:


  ```py
  scanner = ListScanner(lst)
  for value in scanner.safe_iter():
    if some_condition(value):
      value = scanner.advance()
  ```
  ''')



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.contrib.processors.sphinx import generate_sections_markdown
from pydoc_markdown.interfaces import Processor, Resolver
from typing import List, Optional
import docspec
import re


@implements(Processor)
class GoogleProcessor(Struct):
  """
  This class implements the preprocessor for Google and PEP 257 docstrings. It converts
  docstrings formatted in the Google docstyle to Markdown syntax.

  References:

  * https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html
  * https://www.python.org/dev/peps/pep-0257/

  Example:

  ```
  Attributes:
      module_level_variable1 (int): Module level variables may be documented in
          either the ``Attributes`` section of the module docstring, or in an
          inline docstring immediately following the variable.

          Either form is acceptable, but the two should not be mixed. Choose
          one convention to document module level variables and be consistent
          with it.

  Todo:
      * For module TODOs
      * You have to also use ``sphinx.ext.todo`` extension
  ```

  Renders as:

  Attributes:
      module_level_variable1 (int): Module level variables may be documented in
          either the ``Attributes`` section of the module docstring, or in an
          inline docstring immediately following the variable.

          Either form is acceptable, but the two should not be mixed. Choose
          one convention to document module level variables and be consistent
          with it.

  Todo:
      * For module TODOs
      * You have to also use ``sphinx.ext.todo`` extension

  @doc:fmt:google
  """

  _param_res = [
    re.compile(r'^(?P<param>\S+):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+\((?P<type>[^)]+)\):\s+(?P<desc>.+)$'),
    re.compile(r'^(?P<param>\S+)\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{\[(?P<type>\S+)\]\}\s+--\s+(?P<desc>.+)$'),
    re.compile(
      r'^(?P<param>\S+)\s+\{(?P<type>\S+)\}\s+--\s+(?P<desc>.+)$'),
  ]

  _keywords_map = {
    'Args:': 'Arguments',
    'Arguments:': 'Arguments',
    'Attributes:': 'Attributes',
    'Example:': 'Example',
    'Examples:': 'Examples',
    'Keyword Args:': 'Arguments',
    'Keyword Arguments:': 'Arguments',
    'Methods:': 'Methods',
    'Note:': 'Notes',
    'Notes:': 'Notes',
    'Other Parameters:': 'Arguments',
    'Parameters:': 'Arguments',
    'Return:': 'Returns',
    'Returns:': 'Returns',
    'Raises:': 'Raises',
    'References:': 'References',
    'See Also:': 'See Also',
    'Todo:': 'Todo',
    'Warning:': 'Warnings',
    'Warnings:': 'Warnings',
    'Warns:': 'Warns',
    'Yield:': 'Yields',
    'Yields:': 'Yields',
  }

  def check_docstring_format(self, docstring: str) -> bool:
    for section_name in self._keywords_map:
      if section_name in docstring:
        return True
    return False

  @override
  def process(self, modules: List[docspec.Module], resolver: Optional[Resolver]) -> None:
    docspec.visit(modules, self._process)

  def _process(self, node: docspec.ApiObject):
    if not node.docstring:
      return

    lines = []
    current_lines = []
    in_codeblock = False
    keyword = None

    def _commit():
      if keyword:
        generate_sections_markdown(lines, {keyword: current_lines})
      else:
        lines.extend(current_lines)
      current_lines.clear()

    for line in node.docstring.split('\n'):
      if line.startswith("```"):
        in_codeblock = not in_codeblock
        current_lines.append(line)
        continue

      if in_codeblock:
        current_lines.append(line)
        continue

      line = line.strip()
      if line in self._keywords_map:
        _commit()
        keyword = self._keywords_map[line]
        continue

      if keyword is None:
        lines.append(line)
        continue

      for param_re in self._param_res:
        param_match = param_re.match(line)
        if param_match:
          if 'type' in param_match.groupdict():
            current_lines.append(
              '- `{param}` _{type}_ - {desc}'.format(**param_match.groupdict()))
          else:
            current_lines.append(
              '- `{param}` - {desc}'.format(**param_match.groupdict()))
          break

      if not param_match:
        current_lines.append('  {line}'.format(line=line))

    _commit()
    node.docstring = '\n'.join(lines)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Template language for YAML files similar to [YTT][].

  [YTT]: https://get-ytt.io/
"""

from typing import Any, Dict, TextIO, Type, Union
from nr.collections.chaindict import ChainDict
import textwrap
import json
import yaml


def load(
  file_: Union[str, TextIO],
  context: Dict[str, Any],
  Loader: Type[yaml.Loader] = None,
) -> Any:
  """
  Loads a YAML template into a Python object.

  # Arguments
  file_: The file-like object or filename to load.
  context: The context variables available to the template.
  Loader: The #yaml.Loader implementation. Defaults to #yaml.SafeLoader.
  """

  if isinstance(file_, str):
    with open(file_) as fp:
      return load(fp, context, Loader)

  if Loader is None:
    Loader = yaml.SafeLoader

  yaml_code = []
  it = iter(file_)
  for line in it:

    # Parse Python code blocks.
    if line.startswith('#@'):
      block_lines = [line[2:]]
      for line in it:
        if not line.startswith('#@'):
          raise ValueError('missing #@ end')
        if line[2:].strip() == 'end':
          break
        block_lines.append(line[2:])
      code = textwrap.dedent(''.join(block_lines))
      exec(code, context, context)

    # Inline replacements.
    index = line.find('#@')
    if index > 0:
      line = line[:index] + json.dumps(eval(line[index+2:], context, context)) + '\n'

    yaml_code.append(line)

  return yaml.load(''.join(yaml_code), Loader)


class Attributor:

  def __init__(self, data: Dict, default: Any = None) -> None:
    self._data = data
    self._default = default

  def __getattr__(self, name: str) -> Any:
    return self._data.get(name)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from typing import ContextManager, Iterable, IO
import collections
import contextlib
import csv
import hashlib
import nr.fs
import os

FilenameAndHash = collections.namedtuple('FilenameAndHash', 'algorithm,hash,name')


def hash_file(filename: str, algorithm: str, chunksize: int = 2**13) -> str:
  hash_ = hashlib.new(algorithm)
  with open(filename, 'rb') as fp:
    while True:
      data = fp.read(chunksize)
      if not data:
        break
      hash_.update(data)
  return hash_.hexdigest()


class KnownFiles:
  """
  A helper class to keep track of the files that you write so you can get back the
  list of files at a later point.
  """

  def __init__(self, directory: str, filename: str = '.generated-files.txt', hash_algorithm: str = 'md5') -> None:
    self._directory = directory
    self._filename = filename
    self._hash_algorithm = hash_algorithm
    self._files = None

  def __enter__(self) -> 'KnownFiles':
    assert self._files is None, 'Context already entered.'
    self._files = []
    return self

  def __exit__(self, *args) -> None:
    with open(os.path.join(self._directory, self._filename), 'w') as fp:
      writer = csv.writer(fp, delimiter=' ')
      for filename in self._files:
        try:
          hash_ = hash_file(os.path.join(self._directory, filename), self._hash_algorithm)
        except OSError:
          hash_ = '-'
        writer.writerow([self._hash_algorithm, hash_, filename])
    self._files = None

  def load(self) -> Iterable[FilenameAndHash]:
    try:
      with open(os.path.join(self._directory, self._filename), 'r') as fp:
        reader = csv.reader(fp, delimiter=' ')
        for row in reader:
          if len(row) != 3:
            continue
          algorithm, hash_, filename= row
          filename = os.path.join(self._directory, filename)
          yield FilenameAndHash(algorithm, hash_, filename)
    except FileNotFoundError:
      return; yield

  def _check_filename(self, filename: str) -> str:
    filename = os.path.relpath(os.path.abspath(filename), self._directory)
    if not nr.fs.issub(filename):
      raise ValueError('filename must point inside directory {!r}, got {!r}'
                       .format(self._directory, filename))
    return filename

  @contextlib.contextmanager
  def open(self, filename: str, mode: str = 'r', **kwargs) -> ContextManager[IO]:
    filename = self._check_filename(filename)
    if 'w' in mode or 'a' in mode:
      if self._files is None:
        raise RuntimeError('KnownFiles.__enter__() was not called')
      with open(os.path.join(self._directory, filename), mode, **kwargs) as fp:
        yield fp
      self._files.append(filename)

  def append(self, filename: str) -> None:
    filename = self._check_filename(filename)
    open(os.path.join(self._directory, filename), 'r').close()  # Ensure the file exists.
    self._files.append(filename)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.


#: The default configuration that is rendered when using --bootstrap base.
DEFAULT_CONFIG = '''
loaders:
  - type: python
processors:
  - type: filter
  - type: smart
  - type: crossref
renderer:
  type: markdown
'''.lstrip()


#: The default configuration that is rendered when uing --bootstrap mkdocs.
DEFAULT_MKDOCS_CONFIG = '''
loaders:
  - type: python
processors:
  - type: filter
  - type: smart
  - type: crossref
renderer:
  type: mkdocs
  pages:
    - title: Home
      name: index
      source: README.md
    - title: API Documentation
      contents:
        - '*'
  mkdocs_config:
    mkdocs_config:
      site_name: My Project
      theme: readthedocs
'''.lstrip()


#: The default configuration that is rendered when using --bootstrap hugo.
DEFAULT_HUGO_CONFIG = '''
loaders:
  - type: python
processors:
  - type: filter
  - type: smart
  - type: crossref
renderer:
  type: hugo
  config:
    title: My Project
    theme: {clone_url: "https://github.com/alex-shpak/hugo-book.git"}
  # The "book" theme only renders pages in "content/docs" into the nav.
  content_directory: content/docs
  default_preamble: {menu: main}
  pages:
    - title: Home
      name: index
      source: README.md
    - title: API Documentation
      contents:
        - '*'
'''.lstrip()


#: Default configuration for Read the Docs to use Pydoc-Markdown.
READTHEDOCS_FILES = {
  '.readthedocs.yml': '''
version: 2
mkdocs: {}  # tell readthedocs to use mkdocs
python:
  version: 3.7
  install:
  - method: pip
    path: .
  - requirements: .readthedocs-requirements.txt
'''.lstrip(),

  '.readthedocs-custom-steps.yml': '''
steps:
- pydoc-markdown --build --site-dir $SITE_DIR
'''.lstrip(),

  '.readthedocs-requirements.txt': '''
readthedocs-custom-steps
'''.lstrip(),
}



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core.union import (
  IUnionTypeMember,
  IUnionTypeResolver,
  UnknownUnionTypeError,
)
from nr.interface import implements
from nr.pylang.utils import classdef
from typing import Any, Iterable


@implements(IUnionTypeResolver)
class ChainTypeResolver:
  """
  Chain multiple #IUnionTypeResolver instances.
  """

  classdef.comparable('_resolvers')

  def __init__(self, *resolvers: IUnionTypeResolver) -> None:
    self._resolvers = resolvers

  def resolve(self, type_name: str) -> IUnionTypeMember:
    for resolver in self._resolvers:
      try:
        return resolver.resolve(type_name)
      except UnknownUnionTypeError:
        pass
    raise UnknownUnionTypeError(type_name)

  def reverse(self, value: Any) -> IUnionTypeMember:
    for resolver in self._resolvers:
      try:
        return resolver.reverse(value)
      except UnknownUnionTypeError:
        pass
    raise UnknownUnionTypeError(type(value))

  def members(self) -> Iterable[IUnionTypeMember]:
    has_members = False
    for resolver in self._resolvers:
      try:
        yield from resolver.members()
        has_members = True
      except NotImplementedError:
        pass
    if not has_members:
      raise NotImplementedError



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Context, SourceLinker
from typing import List, Optional, Tuple
import docspec
import logging
import os
import nr.fs
import subprocess

logger = logging.getLogger(__name__)


def _getoutput(cmd: List[str], cwd: str = None) -> str:
  logger.debug('running command %r (cwd: %r)', cmd, cwd)
  process = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE)
  stdout = process.communicate()[0].decode()
  if process.returncode != 0:
    raise RuntimeError('process {} exited with non-zero exit code {}'
                       .format(cmd[0], process.returncode))
  return stdout


@implements(SourceLinker)
class GitHubSourceLinker(Struct):
  """
  This is a source linker for code on a GitHub repository. It calculates the URL using
  the source filename relative to the repository root (determined from the current working
  directory) and the ``HEAD`` commit's SHA.
  """

  #: The Git repository owner and name.
  repo = Field(str)

  #: The Github hostname. Defaults to `"github.com"`.
  host = Field(str, default='github.com')

  #: The root directory, relative to the context directory. The context directory
  #: is usually the directory that contains Pydoc-Markdown configuration file.
  #: If not set, the project root will be determined from the Git repository in
  #: the context directory.
  root = Field(str, default=None)

  URL_TEMPLATE = 'https://{host}/{repo}/blob/{sha}/{path}#L{lineno}'

  # SourceLinker

  @override
  def get_source_url(self, obj: docspec.ApiObject) -> str:
    """
    Compute the URL in the GitHub #repo for the API object *obj*.
    """

    if not obj.location:
      return None

    # Compute the path relative to the project root.
    rel_path = os.path.relpath(os.path.abspath(obj.location.filename), self._project_root)
    if not nr.fs.issub(rel_path):
      logger.debug('Ignored API object %s, path points outside of project root.', obj.name)
      return None

    url = self.URL_TEMPLATE.format(
      host=self.host, repo=self.repo, sha=self._sha, path=rel_path, lineno=obj.location.lineno)

    logger.debug('Calculated URL for API object %s is %s', obj.name, url)
    return url

  # PluginBase

  @override
  def init(self, context: Context) -> None:
    """
    This is called before the rendering process. The source linker computes the project
    root and Git SHA at this stage before #get_source_url() is called.
    """

    if self.root:
      self._project_root = os.path.join(context.directory, self.root)
    else:
      self._project_root = _getoutput(['git', 'rev-parse', '--show-toplevel'], cwd=context.directory).strip()
    self._sha = _getoutput(['git', 'rev-parse', 'HEAD'], cwd=self._project_root).strip()
    logger.debug('project_root = %r', self._project_root)
    logger.debug('sha = %r', self._sha)




from . import assert_processor_result
from pydoc_markdown.contrib.processors.sphinx import SphinxProcessor


def test_sphinx_processor(processor=None):
  assert_processor_result(processor or SphinxProcessor(),
  '''
  :param s: A string.
  :param b: An int.
  :return: Something funny.
  ''',
  '''
  **Arguments**:

  - `s`: A string.
  - `b`: An int.

  **Returns**:

  Something funny.
  ''')




from .test_utils import assert_text_equals
from docspec_python import parse_python_module, ParserOptions
from pydoc_markdown import PydocMarkdown
from pydoc_markdown.contrib.processors.filter import FilterProcessor
import io
import textwrap


def assert_code_as_markdown(source_code, markdown, full=False, parser_options=None):
  config = PydocMarkdown()

  # Init the settings in which we want to run the tests.
  config.renderer.insert_header_anchors = False
  config.renderer.add_member_class_prefix = False
  config.renderer.render_toc = False
  config.renderer.render_module_header = full
  filter_processor = next(x for x in config.processors if isinstance(x, FilterProcessor))
  filter_processor.documented_only = False

  # Load the source code as a module.
  modules = [parse_python_module(
    io.StringIO(textwrap.dedent(source_code)),
    filename='<string>',
    module_name='_inline',
    options=parser_options,
  )]

  config.process(modules)
  result = config.renderer.render_to_string(modules)
  assert_text_equals(result, textwrap.dedent(markdown))


def test_preprocessing():
  assert_code_as_markdown(
  '''
  def func(s: str) -> List[str]:
    """ Docstring goes here.

    # Arguments
    s (str): Some string value.

    # Returns
    List[str]: Some more strings. """
  ''',
  '''
  #### func

  ```python
  func(s: str) -> List[str]
  ```

  Docstring goes here.

  __Arguments__

  - __s__ (`str`): Some string value.

  __Returns__

  `List[str]`: Some more strings.
  ''')


def test_starred_arguments():
  assert_code_as_markdown(
  '''
  def a(*args, **kwargs):
      """Docstring goes here."""
  def b(abc, *,):
      """Docstring goes here."""
  def c(abc, *, defg):
      """Docstring goes here."""
  ''',
  '''
  #### a

  ```python
  a(*args, **kwargs)
  ```

  Docstring goes here.

  #### b

  ```python
  b(abc)
  ```

  Docstring goes here.

  #### c

  ```python
  c(abc, *, defg)
  ```

  Docstring goes here.
  ''')


def test_class():
  # https://github.com/NiklasRosenstein/pydoc-markdown/issues/83
  assert_code_as_markdown(
  '''
  class MyError(RuntimeError):
    """ Error raised when my thing happens. """
    pass
  ''',
  '''
  ## MyError Objects

  ```python
  class MyError(RuntimeError)
  ```

  Error raised when my thing happens.
  ''')

  assert_code_as_markdown(
  '''
  class Foo:
    # This is not a member docstring.
    member = None
  ''',
  '''
  ## Foo Objects

  ```python
  class Foo()
  ```

  This is not a member docstring.

  #### member
  ''',
  parser_options=ParserOptions(treat_singleline_comment_blocks_as_docstrings=True))

  assert_code_as_markdown(
  '''
  class Class:
    """
    The class documentation!
    """

    def __init__(self, param):
      """
      The constructor.

      :param param: A parameter.
      """
      self.param = param
  ''',
  '''
  ## Class Objects

  ```python
  class Class()
  ```

  The class documentation!

  #### \\_\\_init\\_\\_

  ```python
   | __init__(param)
  ```

  The constructor.

  **Arguments**:

  - `param`: A parameter.
  ''')


def test_enum():
  assert_code_as_markdown(
  '''
    class PetType(enum.Enum):
      """ Enumeration to identify possible pet types. """
      DOG = 0
      CAT = 1
      MOUSE = 2  #: Mice are rare.
  ''',
  '''
  ## PetType Objects

  ```python
  class PetType(enum.Enum)
  ```

  Enumeration to identify possible pet types.

  #### DOG

  #### CAT

  #### MOUSE
  ''')


def test_module_docstring():
  assert_code_as_markdown(
  '''
  # This is the module docstring.
  ''',
  '''
  # \\_inline

  This is the module docstring.
  ''',
  full=True,
  parser_options=ParserOptions(treat_singleline_comment_blocks_as_docstrings=True))


  assert_code_as_markdown(
  '''
  # LICENSE INFO HERE.

  """ This is the module docstring. """

  ''',
  '''
  # \\_inline

  This is the module docstring.
  ''',
  full=True)


def test_attribute_docstring():
  assert_code_as_markdown(
  '''
  class Foo:
    #: This is a member docstring.
    member = None

    #: This is a member with a type hint.
    int_member: int = 32

    #: A member with type hint but no value.
    float_member: float
  ''',
  '''
  ## Foo Objects

  ```python
  class Foo()
  ```

  #### member

  This is a member docstring.

  #### int\\_member

  This is a member with a type hint.

  #### float\\_member

  A member with type hint but no value.
  ''')

  assert_code_as_markdown(
  '''
  class Foo:
    member = None
    """ This is a member docstring. """
  ''',
  '''
  ## Foo Objects

  ```python
  class Foo()
  ```

  #### member

  This is a member docstring.
  ''')



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Loads Python source code.
"""

import docspec
import docspec_python
import logging
import os
import sys

from nr.databind.core import Field, Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Context, Loader, LoaderError
from typing import Iterable, List

logger = logging.getLogger(__name__)


@implements(Loader)
class PythonLoader(Struct):
  """
  This implementation of the #Loader interface parses Python modules and packages using
  #docspec_python. See the options below to control which modules and packages are being
  loaded and how to configure the parser.

  With no #modules or #packages set, the #PythonLoader will discover available modules
  in the current and `src/` directory.

  __lib2to3 Quirks__

  Pydoc-Markdown doesn't execute your Python code but instead relies on the
  `lib2to3` parser. This means it also inherits any quirks of `lib2to3`.

  _List of known quirks_

  * A function argument in Python 3 cannot be called `print` even though
    it is legal syntax
  """

  #: A list of module names that this loader will search for and then parse.
  #: The modules are searched using the #sys.path of the current Python
  # interpreter, unless the #search_path option is specified.
  modules = Field([str], default=None)

  #: A list of package names that this loader will search for and then parse,
  #: including all sub-packages and modules.
  packages = Field([str], default=None)

  #: The module search path. If not specified, the current #sys.path is
  #: used instead. If any of the elements contain a `*` (star) symbol, it
  #: will be expanded with #sys.path.
  search_path = Field([str], default=None)

  #: List of modules to ignore when using module discovery on the #search_path.
  ignore_when_discovered = Field([str], default=lambda: ['test', 'tests', 'setup'])

  #: Options for the Python parser.
  parser = Field(docspec_python.ParserOptions, default=Field.DEFAULT_CONSTRUCT)

  _context = Field(Context, default=None, hidden=True)

  def get_effective_search_path(self) -> List[str]:
    if self.search_path is None:
      search_path = ['.', 'src'] if self.modules is None else list(sys.path)
    else:
      search_path = list(self.search_path)
      if '*' in search_path:
        index = search_path.index('*')
        search_path[index:index+1] = sys.path
    return [os.path.join(self._context.directory, x) for x in search_path]

  # Loader

  @override
  def load(self) -> Iterable[docspec.Module]:
    search_path = self.get_effective_search_path()
    modules = list(self.modules or [])
    packages = list(self.packages or [])
    do_discover = (self.modules is None and self.packages is None)

    if do_discover:
      for path in search_path:
        try:
          discovered_items = list(docspec_python.discover(path))
        except FileNotFoundError:
          continue

        logger.info(
          'Discovered Python modules %s and packages %s in %r.',
          [x.name for x in discovered_items if x.is_module()],
          [x.name for x in discovered_items if x.is_package()],
          path,
        )

        for item in discovered_items:
          if item.name in self.ignore_when_discovered:
            continue
          if item.is_module():
            modules.append(item.name)
          elif item.is_package():
            packages.append(item.name)

    logger.info(
      'Load Python modules (search_path: %r, modules: %r, packages: %r, discover: %s)',
      search_path, modules, packages, do_discover
    )

    return docspec_python.load_python_modules(
      modules=modules,
      packages=packages,
      search_path=search_path,
      options=self.parser
    )

  # PluginBase

  @override
  def init(self, context: Context) -> None:
    self._context = context




from . import test_google, test_pydocmd, test_sphinx
from pydoc_markdown.contrib.processors.smart import SmartProcessor


def test_google_style():
  test_google.test_google_processor(SmartProcessor())


def test_sphinx_style():
  test_sphinx.test_sphinx_processor(SmartProcessor())


def test_pydocmd_style():
  test_pydocmd.test_pydocmd_processor(SmartProcessor())



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, ProxyType, Remainder, Struct
from nr.interface import implements, override
from pydoc_markdown.contrib.renderers.markdown import MarkdownRenderer
from pydoc_markdown.interfaces import Context, Renderer, Resolver, Server, Builder
from pydoc_markdown.util.knownfiles import KnownFiles
from pydoc_markdown.util.pages import Page
from urllib.parse import urlparse, urljoin
from typing import Dict, Iterable, List, Optional, TextIO
import docspec
import logging
import nr.fs
import os
import platform as _platform
import posixpath
import requests
import shutil
import subprocess
import sys
import tarfile
import toml
import yaml

logger = logging.getLogger(__name__)
HugoPage = ProxyType()


@HugoPage.implementation
class HugoPage(Page):
  """
  A subclass of #Page which adds Hugo-specific overrides.

  ### Options
  """

  children = Field([HugoPage], default=list)

  #: The Hugo preamble of the page. This is merged with the #HugoRenderer.default_preamble.
  preamble = Field(dict, default=dict)

  #: Override the directory that this page is rendered into (relative to the
  #: content directory). Defaults to `null`.
  directory = Field(str, default=None)


class HugoThemePath(Struct):
  path = Field(str)

  @property
  def name(self):
    return os.path.basename(self.path)

  def install(self, themes_dir: str):
    # TODO (@NiklasRosenstein): Support Windows (which does not support symlinking).
    dst = os.path.join(self.name)
    if not os.path.lexists(dst):
      os.symlink(self.path, dst)


class HugoThemeGitUrl(Struct):
  clone_url = Field(str)
  postinstall = Field([str], default=list)

  @property
  def name(self):
    return posixpath.basename(self.clone_url).rstrip('.git').lstrip('hugo-')

  def install(self, theme_dir: str):
    dst = os.path.join(theme_dir, self.name)
    if not os.path.isdir(dst):
      command = ['git', 'clone', self.clone_url, dst, '--depth', '1',
                 '--recursive', '--shallow-submodules']
      subprocess.check_call(command)
      for command in self.postinstall:
        subprocess.check_call(command, shell=True, cwd=dst)


class HugoConfig(Struct):
  """
  Represents the Hugo configuration file that is rendered into the build directory.

  ### Options
  """

  #: Base URL.
  baseURL = Field(str, default=None)

  #: Language code. Default: `en-us`
  languageCode = Field(str, default='en-us')

  #: Title of the site. This is a mandatory field.
  title = Field(str)

  #: The theme of the site. This is a mandatory field. It must be a string, a #HugoThemePath
  #: or a #HugoThemeGitUrl object. Examples:
  #:
  #: ```yml
  #: theme: antarctica
  #: theme: {clone_url: "https://github.com/alex-shpak/hugo-book.git"}
  #: theme: docs/hugo-theme/
  #: ```
  theme = Field((str, HugoThemePath, HugoThemeGitUrl))

  #: This field collects all remaining options that do not match any of the above
  #: and will be forwarded directly into the Hugo `config.yaml` when it is rendered
  #: into the build directory.
  additional_options = Field(dict, Remainder())

  def to_toml(self, fp: TextIO) -> None:
    data = self.additional_options.copy()
    for field in self.__fields__:
      if field in ('additional_options', 'theme'):
        continue
      value = getattr(self, field)
      if value:
        data[field] = value
    if isinstance(self.theme, str):
      data['theme'] = self.theme
    elif isinstance(self.theme, (HugoThemePath, HugoThemeGitUrl)):
      data['theme'] = self.theme.name
    else: assert False
    fp.write(toml.dumps(data))


@implements(Renderer, Server, Builder)
class HugoRenderer(Struct):
  """
  A renderer that produces Markdown files compatible with [Hugo][0]. The `--bootstrap hugo`
  option can be used to create a Pydoc-Markdown configuration file with the Hugo template.

  * Adds a YAML preamble to every generated Markdown file.
  * Produces files in a layout suitable for Hugo (e.g. `_index.md` files).
  * Produces a `config.yaml` if #config is not set to `null`.
  * Can be used with the Pydoc-Markdown `--server` option to live-preview the generated docs.
  * Downloads a suitable Hugo binary from Github if Hugo is not already installed (see #get_hugo).

  Example configuration:

  ```yml
  renderer:
    type: hugo
    config:
      title: My Project
      theme: {clone_url: "https://github.com/alex-shpak/hugo-book.git"}
    # The "book" theme only renders pages in "content/docs" into the nav.
    content_directory: content/docs
    default_preamble: {menu: main}
    pages:
      - title: Home
        name: index
        source: README.md
      - title: API Documentation
        contents:
          - '*'
  ```

  [0]: https://gohugo.io/

  ### Options
  """

  #: The directory where all generated files are placed. Default: `build/docs`
  build_directory = Field(str, default='build/docs')

  #: The directory _inside_ the build directory where the generated
  #: pages are written to. Default: `content`
  content_directory = Field(str, default='content')

  #: Clean up files that were previously generated by the renderer before the next
  #: render pass. Defaults to `True`.
  clean_render = Field(bool, default=True)

  #: The pages to render.
  pages = Field(HugoPage.collection_type())

  #: The default Hugo preamble that is applied to every page. Example:
  #:
  #: ```yml
  #: default_preamble:
  #:   menu: main
  #: ```
  default_preamble = Field(dict, default=dict)

  #: The #MarkdownRenderer configuration.
  markdown = Field(MarkdownRenderer, default=Field.DEFAULT_CONSTRUCT)

  #: The contents of the Hugo `config.toml` file as YAML. This can be set to `null` in
  #: order to not produce the `config.toml` file in the #build_directory. Must be deserializable
  #: into a #HugoConfig.
  config = Field(HugoConfig, nullable=True)

  #: Options for when the Hugo binary is not present and should be downloaded
  #: automatically. Example:
  #:
  #: ```yml
  #: get_hugo:
  #:   enabled: true
  #:   version: '0.71'
  #:   extended: true
  #: ```
  get_hugo = Field({
    'enabled': Field(bool, default=True),
    'version': Field(str, default=None),
    'extended': Field(bool, default=True),
  }, default=Field.DEFAULT_CONSTRUCT)

  _context = Field(Context, default=None, hidden=True)  # Initialized in #init()

  def _render_page(self, modules: List[docspec.Module], page: HugoPage, filename: str):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    preamble = dict(**self.default_preamble, **{'title': page.title}, **page.preamble)

    with open(filename, 'w') as fp:
      fp.write('---\n')
      fp.write(yaml.safe_dump(preamble))
      fp.write('---\n\n')

      if page.source:
        with open(os.path.join(self._context.directory, page.source)) as src:
          shutil.copyfileobj(src, fp)
      else:
        self.markdown.fp = fp
        self.markdown.render(modules)
        self.markdown.fp = None

  def _get_hugo_bin(self):
    hugo_bin = shutil.which('hugo')
    if not hugo_bin and self.get_hugo.enabled:
      hugo_bin = os.path.abspath(os.path.join(self.build_directory, '.bin', 'hugo'))
      if not os.path.isfile(hugo_bin):
        install_hugo(hugo_bin, self.get_hugo.version, self.get_hugo.extended)
    if not hugo_bin:
      raise RuntimeError('Hugo is not installed')
    return hugo_bin

  # Renderer

  @override
  def render(self, modules: List[docspec.Module]) -> None:
    known_files = KnownFiles(self.build_directory)
    content_dir = os.path.join(self.build_directory, self.content_directory)

    if self.clean_render:
      for file_ in known_files.load():
        try:
          os.remove(file_.name)
        except FileNotFoundError as exc:
          pass

    # Render the pages.
    with known_files:
      for item in self.pages.iter_hierarchy():
        if item.page.name == 'index':
          item.page.name = '_index'
        page_content_dir = content_dir
        if item.page.directory:
          page_content_dir = os.path.normpath(os.path.join(content_dir, item.page.directory))
        filename = item.filename(page_content_dir, '.md', index_name='_index', skip_empty_pages=False)
        self._render_page(item.page.filtered_modules(modules), item.page, filename)
        known_files.append(filename)

      # Render the config file.
      if self.config is not None:
        if isinstance(self.config.theme, (HugoThemePath, HugoThemeGitUrl)):
          self.config.theme.install(os.path.join(self.build_directory, 'themes'))
        filename = os.path.join(self.build_directory, 'config.toml')
        logger.info('Rendering "%s"', filename)
        with known_files.open(filename, 'w') as fp:
          self.config.to_toml(fp)

  @override
  def get_resolver(self, modules: List[docspec.Module]) -> Optional[Resolver]:
    # TODO (@NiklasRosenstein): The resolver returned by the Markdown
    #   renderer does not implement linking across multiple pages.
    return self.markdown.get_resolver(modules)

  # Server

  @override
  def get_server_url(self) -> str:
    urlinfo = urlparse(self.config.baseURL or '')
    return urljoin('http://localhost:1313/', urlinfo.path)

  @override
  def start_server(self) -> subprocess.Popen:
    hugo_bin = self._get_hugo_bin()
    command = [hugo_bin, 'server']
    logger.info('Running %s in "%s"', command, self.build_directory)
    return subprocess.Popen(command, cwd=self.build_directory)

  # Builder

  @override
  def build(self, site_dir: str=None) -> None:
    command = [self._get_hugo_bin()]
    if site_dir:
      command += ['--destination', os.path.abspath(site_dir)]
    subprocess.check_call(command, cwd=self.build_directory)

  # PluginBase

  @override
  def init(self, context: Context) -> None:
    self._context = context
    self.markdown.init(context)


def install_hugo(to: str, version: str = None, extended: bool = True) -> None:
  """
  Downloads the latest release of *Hugo* from [Github](https://github.com/gohugoio/hugo/releases)
  and places it at the path specified by *to*. This will install the extended version if it is
  available and *extended* is set to `True`.

  :param to: The file to write the Hugo binary to.
  :param version: The Hugo version to get. If not specified, the latest release will be used.
  :param extended: Whether to download the "Hugo extended" version. Defaults to True.
  """

  # TODO (@NiklasRosenstein): Support BSD platforms.

  if sys.platform.startswith('linux'):
    platform = 'Linux'
  elif sys.platform.startswith('win32'):
    platform = 'Windows'
  elif sys.platform.startswith('darwin'):
    platform = 'macOS'
  else:
    raise EnvironmentError('unsure how to get a Hugo binary for platform {!r}'.format(sys.platform))

  machine = _platform.machine().lower()
  if machine in ('x86_64', 'amd64'):
    arch = '64bit'
  elif machine in ('i386',):
    arch = '32bit'
  else:
    raise EnvironmentError('unsure whether to intepret {!r} as 32- or 64-bit.'.format(machine))

  releases = get_github_releases('gohugoio/hugo')
  if version:
    version = version.lstrip('v')
    for release in releases:
      if release['tag_name'].lstrip('v') == version:
        break
    else:
      raise ValueError('no Hugo release for version {!r} found'.format(version))
  else:
    release = next(releases)
    version = release['tag_name'].lstrip('v')

  files = {asset['name']: asset['browser_download_url'] for asset in release['assets']}

  hugo_archive = 'hugo_{}_{}-{}.tar.gz'.format(version, platform, arch)
  hugo_extended_archive = 'hugo_extended_{}_{}-{}.tar.gz'.format(version, platform, arch)
  if extended and hugo_extended_archive in files:
    filename = hugo_extended_archive
  else:
    filename = hugo_archive

  logger.info('Downloading Hugo v%s from "%s"', version, files[filename])
  os.makedirs(os.path.dirname(to), exist_ok=True)
  with nr.fs.tempdir() as tempdir:
    path = os.path.join(tempdir.name, filename)
    with open(path, 'wb') as fp:
      shutil.copyfileobj(requests.get(files[filename], stream=True).raw, fp)
    with tarfile.open(path) as archive:
      with open(to, 'wb') as fp:
        shutil.copyfileobj(archive.extractfile('hugo'), fp)

  nr.fs.chmod(to, '+x')
  logger.info('Hugo v%s installed to "%s"', version, to)


def get_github_releases(repo: str) -> Iterable[dict]:
  """
  Returns an iterator for all releases of a Github repository.
  """

  url = 'https://api.github.com/repos/{}/releases'.format(repo)
  while url:
    response = requests.get(url)
    links = parse_links_header(response.headers.get('Link'))
    url = links.get('next')
    yield from response.json()


def parse_links_header(link_header: str) -> Dict[str, str]:
  """
  Parses the `Link` HTTP header and returns a map of the links. Logic from
  [PageLinks.java](https://github.com/eclipse/egit-github/blob/master/org.eclipse.egit.github.core/src/org/eclipse/egit/github/core/client/PageLinks.java#L43-75).
  """

  links = {}

  for link in link_header.split(','):
    segments = link.split(';')
    if len(segments) < 2:
      continue
    link_part = segments[0].strip()
    if not link_part.startswith('<') or not link_part.endswith('>'):
      continue
    link_part = link_part[1:-1]
    for rel in (x.strip().split('=') for x in segments[1:]):
      if len(rel) < 2 or rel[0] != 'rel':
        continue
      rel_value = rel[1]
      if rel_value.startswith('"') and rel_value.endswith('"'):
        rel_value = rel_value[1:-1]
      links[rel_value] = link_part

  return links




from pydoc_markdown.test.test_utils import assert_text_equals
import docspec
import textwrap


def assert_processor_result(processor, docstring, expected_output):
  module = docspec.Module(name='test', location=None, docstring=textwrap.dedent(docstring), members=[])
  processor.process([module], None)
  assert_text_equals(module.docstring, textwrap.dedent(expected_output))



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Field, Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor, Resolver
from pydoc_markdown.contrib.processors.google import GoogleProcessor
from pydoc_markdown.contrib.processors.pydocmd import PydocmdProcessor
from pydoc_markdown.contrib.processors.sphinx import SphinxProcessor
from typing import List, Optional
import docspec


@implements(Processor)
class SmartProcessor(Struct):
  """
  This processor picks the #GoogleProcessor, #SphinxProcessor or #PydocmdProcessor after
  guessing which is appropriate from the syntax it finds in the docstring.
  """

  google = Field(GoogleProcessor, default=GoogleProcessor)
  pydocmd = Field(PydocmdProcessor, default=PydocmdProcessor)
  sphinx = Field(SphinxProcessor, default=SphinxProcessor)

  @override
  def process(self, modules: List[docspec.Module], resolver: Optional[Resolver]) -> None:
    docspec.visit(modules, self._process)

  def _process(self, obj: docspec.ApiObject):
    if not obj.docstring:
      return None

    for name in ('google', 'pydocmd', 'sphinx'):
      indicator = '@doc:fmt:' + name
      if indicator in obj.docstring:
        obj.docstring = obj.docstring.replace(indicator, '')
        return getattr(self, name)._process(obj)

    if self.sphinx.check_docstring_format(obj.docstring):
      return self.sphinx._process(obj)
    if self.google.check_docstring_format(obj.docstring):
      return self.google._process(obj)
    return self.pydocmd._process(obj)



# -*- coding: utf8 -*-
# Copyright (c) 2019 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.databind.core import Struct
from nr.interface import implements, override
from pydoc_markdown.interfaces import Processor, Resolver
from typing import Dict, List, Optional
import docspec
import logging
import re

logger = logging.getLogger(__name__)


@implements(Processor)
class CrossrefProcessor(Struct):
  """
  Finds references to other objects in Markdown docstrings and produces links to other
  pages. The links are provided by the current #Renderer via the #Resolver interface.

  > __Note__: This processor is a work in progress, and most of the time it just converts
  > references into inline-code.

  The syntax for cross references is as follows:

  ```
  This is a ref to another class: #PydocmdProcessor
  You can rename a ref like #this~PydocmdProcessor
  And you can append to the ref name like this: #PydocmdProcessor#s
  ```

  Renders as

  > This is a ref to another class: #PydocmdProcessor
  > You can rename a ref like #this~PydocmdProcessor
  > And you can append to the ref name like this: #PydocmdProcessor#s

  Example configuration:

  ```yml
  processors:
    - type: crossref
  ```
  """

  @override
  def process(self, modules: List[docspec.Module], resolver: Optional[Resolver]):
    unresolved = {}
    if resolver:
      reverse = docspec.ReverseMap(modules)
      docspec.visit(modules, lambda x: self._preprocess_refs(x, resolver, reverse, unresolved))

    if unresolved:
      summary = []
      for uid, refs in unresolved.items():
        summary.append('  {}: {}'.format(uid, ', '.join(refs)))

      logger.warning(
        '%s cross-reference(s) could not be resolved:\n%s',
        sum(map(len, unresolved.values())),
        '\n'.join(summary),
      )

  def _preprocess_refs(
    self,
    node: docspec.ApiObject,
    resolver: Resolver,
    reverse: docspec.ReverseMap,
    unresolved: Dict[str, List[str]],
  ) -> None:
    if not node.docstring:
      return

    def handler(match):
      ref = match.group('ref')
      parens = match.group('parens') or ''
      trailing = (match.group('trailing') or '').lstrip('#')
      # Remove the dot from the ref if its trailing (it is probably just
      # the end of the sentence).
      has_trailing_dot = False
      if trailing and trailing.endswith('.'):
        trailing = trailing[:-1]
        has_trailing_dot = True
      elif not parens and ref.endswith('.'):
        ref = ref[:-1]
        has_trailing_dot = True
      href = resolver.resolve_ref(node, ref)
      if href:
        result = '[`{}`]({})'.format(ref + parens + trailing, href)
      else:
        uid = '.'.join(x.name for x in reverse.path(node))
        unresolved.setdefault(uid, []).append(ref)
        result = '`{}`'.format(ref + parens)
      # Add back the dot.
      if has_trailing_dot:
        result += '.'
      return result

    node.docstring = re.sub(
      r'\B#(?P<ref>[\w\d\._]+)(?P<parens>\(\))?(?P<trailing>#[\w\d\._]+)?',
      handler,
      node.docstring)
